/*
 Time-stamp: "2011-01-07 14:57:20 yamagen"

 jk -- j[apanese] k[ey word in context] display 
 Copyright (c) 2000-2011 Hilofumi Yamamoto <yamagen@ucsd.edu>

 This file may be freely modified/redistributed.

 Thanks Prof. Haruhiko Okumura <okumura@matsusaka-u.ac.jp>
 (ukanji.c,mgrep.c)

 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <regex.h>
#include <locale.h>
#include "config.h"

int digitstr(char *str);
void usage(void);
int iskanji(unsigned char c,int code);
int find_key(char *linebuf, int linelen, char *key, int keylen, int pos);
int find_key_regex(char *linebuf, int linelen, char *key, int keylen, int pos);
int kwic(FILE *fp, char *key,char *fname);
int print_key(char *linebuf, int pos, char *key);
int print_nchar(char *buf, int n, int col, char *key);
int main (int argc, char *argv[]);
regex_t compile_pattern(char *key);

#define	K1_C	' '   /* line end, half kanji */
#define	K2_C	' '   /* line begin, half kanji */
#define	PRE_C	' '
#define	TAB_C	' '
#define	LINE_W	80
#define	L_WING	36
#define	BUF	4096
#define	TRUE	1
#define	FALSE	0
#define	ANCTAG	0xffffffff

/* flags */
int code = 0;
int anchor_on = 0;
int regex_on = 0;
int fileon = 0;
int file_only = 0;
int wing_size = L_WING;
int line_width = LINE_W;
int delimiter = 0;
int right_only = 0;
int left_only = 0;

int digitstr(char *str)
{
    if (*str == '\0')	return FALSE;
    for (; *str != '\0'; str++) {
	if (!isdigit(*str))	return FALSE;
    }
    return TRUE;
}

void usage(void)
{
  fprintf(stderr,
	  "Usage : %% %s pattern [-adfFhlrsV] [-Wnum] [-Lnum] [file ...]\n"
	  "option:\n"
	  "     -a shows anchor tag embeded between '0xff'.\n"
	  "     -e use regular expression.\n"
	  "     -d shows keyword between two delimiters `|'.\n"
	  "     -f shows filename and line.\n"
	  "     -F shows filename only.\n"
	  "     -h shows this help. \n"
	  "     -l shows only the left wing. \n"
	  "     -r shows only the right wing. \n"
#ifdef SJIS
	  "     -s processes euc file.(default=sjis)\n"
#else 
	  "     -s processes sjis file.(default=euc)\n"
#endif
	  "     -V shows version.\n"
	  "     -Wnum set num as the number of the left wing width.\n"
	  "           num should be set by byte (default=36).\n"
	  "     -Lnum set num as the number of the line width.\n"
	  "           num should be set by byte (default=80).\n"
	  ,PACKAGE
	  );
  exit(0);
}

int iskanji(unsigned char c, int sjis)
{
#ifdef SJIS
    if (sjis) 
	sjis = 0;
#else
    if (sjis) 
	sjis = 1;
#endif
    if (sjis)
	return((0x81 <= c && c <= 0x9F) || (0xE0 <= c && c <= 0xFC)); /* SJIS */
    else 
	return(0x8E <= c && c <= 0xFE && 0x8F != c && c != 0xA0);     /* EUC  */
}

int find_key(char *linebuf, int linelen, char *key, int keylen, int pos)
{
    while (1) {
	if (pos + keylen > linelen) 
	    return(-1);
	else 
	    if (strncmp(linebuf + pos, key, keylen) == 0) 
		return(pos);
	    else 
		if (iskanji(linebuf[pos], code)) 
		    pos += 2;
		else 
		    pos++;
    }
}


int print_nchar(char *buf, int n,int col,char *key)
{
  int i, j = 0, kanjiin = 0;
  int keylen = 0;
  keylen = strlen(key);

  if (right_only) {
    j = keylen;
  }

  for (i = j; i < n; i++) {
    if(delimiter && (i + col == wing_size || i + col == wing_size + keylen)){
      putchar('|'); /* delimiter option */
    }
    if (kanjiin) {
      putchar(buf[i]);
      kanjiin = 0;
    } else if (iskanji(buf[i],code)) {
      if (i == n - 1) {
	/* half of kanji char will not be displayed. */
	putchar(K1_C);
	return 0;
      } else {
	putchar(buf[i]);
	kanjiin = 1;
      }
    } else if (buf[i] == '\t') {
      putchar(TAB_C);
    } else if (buf[i] == '\n') {
      return 0;
    } else {
      putchar(buf[i]);
    }
  }
  return 0;
}

int print_key(char *linebuf, int pos, char *key)
{
  int i, kanjiin, column = 0;
  int spaces = wing_size - pos;

  if (spaces > 0) {
    while (spaces-- > 0) {	/* put spaces at the begining of the line */
      putchar(PRE_C);
      column++;
    }
    print_nchar(linebuf, line_width - column, column, key);
  } else {			/* delete chars at the begining of the line */
    kanjiin = 0;
    for (i = 0; spaces++ < 0; i++) {
      if (kanjiin) {
	kanjiin = 0;
      } else if (iskanji(linebuf[i],code)) {
	kanjiin = 1;
      }
    }
    if (kanjiin) {
      putchar(K2_C);
      column++;		/* mojibake protect */
      i++;
    }
    print_nchar(linebuf + i, line_width - column, column, key);
  }
  putchar('\n');
  return 0;
}

int kwic(FILE *fp, char *key, char *fname)
{
    int pos,buflen,keylen = strlen(key),c,i=0;
    long line = 1L;
    char buf[BUF];
    char anchor[BUF];

    while (fgets(buf, BUF, fp) != NULL) {
	pos = 0;
	buflen = strlen(buf);

	/* anchor tag */
	if((c = buf[pos] ) == ANCTAG){
	    i = 0;
	    while( (c = buf[++pos] ) != ANCTAG ) {
		anchor[i++] = c; 
	    }
	}
		
/* 	while ((pos = find_key(buf, buflen, key, keylen, pos)) >= 0) {  */

	while ((pos = find_key_regex(buf, buflen, key, keylen, pos)) >= 0) { 
	    if (anchor_on) {
		printf("%s:",anchor);
	    }
	    if (fileon) {
		printf("%s %5ld:", fname, line);
	    }
	    if (file_only) {
		printf("%s", fname);
	    }
	    print_key(buf, pos, key);
	    pos += keylen;
	}
	line++;
    }
    return 0;
}

int main (int argc, char **argv)
{
    char *format;
    char *op;
    char opt_show = 'b';
    int pattern_given = 0;
    int files_processed = 0;
    char *key = 0;
    FILE *fp;
    regex_t reg;


    setlocale(LC_ALL, "");

    for (argv++; *argv && argv[0][0] == '-' && argv[0][1]; argv++) {
	for (op = &argv[0][1]; *op; op++) {
	    switch (*op) {
	    case 'a': anchor_on = 1; break;
	    case 'd': delimiter = 1; break;
	    case 'e': regex_on = 1; break;
	    case 'f': fileon = 1;    break;
	    case 'F': file_only = 1; break;
	    case 's': code = 1;    break;
	    case 't': opt_show = *op; break; /* for test */
	    case 'l': left_only = 1; right_only = 0; break;
	    case 'r': right_only = 1; left_only = 0; break;
	    case 'L':
		if (digitstr(op + 1) == FALSE) usage();
		line_width = atoi(op + 1);
		/*	printf("num of line width === %d\n",line_width); */
		while(*(++op + 1))
		    ;
		break;
	    case 'W':
		if (digitstr(op + 1) == FALSE)  usage();
		wing_size = atoi(op + 1);
		/*	printf("num of wing size === %d\n",wing_size); */
		while(*(++op + 1))
		    ;
		break;
	    case 'V': 
		fprintf(stderr,"%s version %s\n", PACKAGE, VERSION);
		exit(0); break;
	    case 'h': usage();
	    default:
		fprintf(stderr,"invalid option -- %c\n", *op);
		usage();
		exit(0);
	    }
	} 
    }

    if (delimiter && !right_only && !left_only)
	line_width -= 2;

    if (pattern_given == 0) {	
	if(*argv){
	    key = *argv;
	    pattern_given = 1;
	    /*      printf("Ok! Your request string is %s.\n",*argv); */
	}
/* 	reg = compile_pattern(key); */
    } 

    if (right_only) {
	delimiter = 0;
	wing_size = 0; 
	line_width += strlen(key);
    }
    if (left_only) {
	delimiter = 0;
	wing_size = line_width;
    }
    
    if (file_only) {
	wing_size = line_width = delimiter = 0;
    }

    for(*argv++; *argv; *argv++){
	if (pattern_given) {
	    if ((fp = fopen(*argv, "r")) == NULL) {
		fprintf(stderr, "Can't open %s\n", *argv);
		exit(1);
	    }
	    kwic(fp, key, *argv);
	    fclose(fp);
	    files_processed++;
	}
    }
    
    if (pattern_given == 0)
	usage();
    if (files_processed == 0)
	kwic(stdin, key, "stdin");
    
    return 0;
}


regex_t compile_pattern(char *key)
{
    regex_t reg;
    int ret = 0;
    char err[100];

    ret = regcomp(&reg, key, REG_EXTENDED);
    if (ret != 0) {
	regerror(ret, &reg, err, sizeof err);
	printf("regcomp error: %s\n", err);
	regfree(&reg);
	exit(0);
    }
    return reg;
}

int find_key_regex(char *linebuf, int linelen, char *key, int keylen, int pos)
{

    regex_t reg;
    int ret = 0;
    char err[100];
    regmatch_t match[0];

/*     int buf_size = sizeof linebuf / sizeof linebuf[0]; */

    ret = regcomp(&reg, key, REG_EXTENDED);
    if (ret != 0) {
	regerror(ret, &reg, err, sizeof err);
	printf("regcomp error: %s\n", err);
	regfree(&reg);
	exit(0);
    }

    while (1) {
	if (pos + keylen > linelen) 
	    return(-1);
	else 
/*  	    if (strncmp(linebuf + pos, key, keylen) == 0)   */
	    if(regexec(&reg, linebuf + pos, 1, match, 0) != REG_NOMATCH) {
/* 		printf("so %d eo %d\n", match[0].rm_so,match[0].rm_eo); */
/* 		strncpy(buf, linebuf + match[0].rm_so, 30); */
/* 		printf("%s\n", buf); */
/* 		return(pos); */
 		return(match[0].rm_so); 
	    }
	    else  
		if (iskanji(linebuf[pos], code))
		    pos += 2;
		else
		    pos++;
    }
    regfree(&reg);
}

